--char.e
--draw characters on graphics screen
--by David Money, July, 2006

include graphx.e
integer color,bold,italic,accent
atom scale,width,end_x,end_y
bold=0 italic=0 color=14
scale=1 width=8 end_x=0 end_y=5
accent=0 --vertical offset for accent marks

global procedure text_attributes(atom txt_scale,integer txt_color,integer txt_bold,
	    integer txt_italic)
    --set scale, color, bold, and italics for graphics_text() or continue_text()
    scale=txt_scale color=txt_color
    if txt_bold then
	bold=1
    else
	bold=0
    end if
    if txt_italic then
	italic=1
    else
	italic=0
    end if
end procedure

procedure graphics_info()
    --abort if rectangle() or graphics_text() is called in text mode
    sequence config
    config=machine_func(13,0)
    hres=config[5]
    vres=config[6]
    if not hres then
	printf(1,"Mode %d is text only -- cannot do graphics characters.",{config[2]})
	abort(1)
    end if
end procedure

global procedure rectangle(integer bkgnd_color,sequence corner1,sequence corner2)
    --change background to specified color in rectangular area; coordinates of
    --corners can be from 0 to 1, or actual pixels
    atom x1,x2,y1,y2,temp
    graphics_info()
    x1=corner1[1] y1=corner1[2]
    x2=corner2[1] y2=corner2[2]
    if (x1+y1)<=2 then
	x1*=hres y1*=vres
    end if
    if (x2+y2)<=2 then
	x2*=hres y2*=vres
    end if
    if x1>x2 then
	temp=x1 x1=x2 x2=temp
    end if
    if y1>y2 then
	temp=y1 y1=y2 y2=temp
    end if
    x1=floor(x1) y1=floor(y1)
    x2=floor(x2) y2=floor(y2)
    machine_proc(11,{bkgnd_color,1,{{x1,y1},{x2,y1},{x2,y2},{x1,y2},{x1,y1}}})
end procedure

procedure draw_modified_path(sequence path,sequence pos)
    --draw path given in graphics_char, offset by pos and multiplied by scale
    for n=1 to length(path) do
	for pt=1 to length(path[n]) do
	    if italic then
		path[n][pt][1]-=path[n][pt][2]*.4 --adjust italic angle
	    end if
	    path[n][pt]*=scale
	    path[n][pt]+=pos
	end for
    end for
    for n=1 to length(path) do
	machine_proc(2,{color,0,path[n]})
    end for
end procedure

procedure draw_dots(sequence dots,sequence pos)
    --draw dot in character like : or i; single pixel if scale<=1; otherwise,
    --4 pixels
    for n=1 to length(dots) do
	if italic then
	    dots[n][1]-=dots[n][2]*.4
	end if
	dots[n]*=scale
	dots[n]+=pos
	pixel(color,dots[n])
	if scale>1 then
	    pixel(color,dots[n]+{-1,0})
	    pixel(color,dots[n]+{0,-1})
	    pixel(color,dots[n]+{-1,-1})
	end if
    end for
end procedure

procedure graphics_char(integer ch,sequence pos)
    --list points to be connected to draw character; center of char is at (0,0);
    --horiz will be from -3 to +3, vertical from -4 to +4 for full-size chars
    --like H; narrow chars are left justified in area, and then the area is
    --shifted (fixed spacing), or the next char is started sooner (proportional)
    integer dx
    sequence path,dots
    accent=0
    dx=8 --char width; for reference; not returned
    path={} dots={}
    if ch>280 or ch<32 then
	ch=0
    end if
    if ch=0 then
	path={{{-3,-4},{3,-4},{3,4},{-3,4},{-3,-4}}}
    elsif ch=32 then
	path={} dx=5
    elsif ch='!' then
	path={{{-2,-4},{-2,1}},{{-2,4},{-2,4}}} dx=4
    elsif ch='\"' then
	path={{{-2,-4},{-2,-2}},{{0,-4},{0,-2}}} dx=6
    elsif ch='#' then
	path={{{-2,-3},{-2,3}},{{2,-3},{2,3}},{{-3,-2},{3,-2}},{{-3,2},{3,2}}}
    elsif ch='$' then
	path={{{3,-2},{0,-3},{-3,-2},{-3,-1},{3,1},{3,2},{0,3},{-3,2}},{{-1,-4},{-1,4}},{{1,-4},{1,4}}}
    elsif ch='%' then
	path={{{-2,-4},{-3,-3},{-2,-2},{-1,-3},{-2,-4}},{{2,2},{1,3},{2,4},{3,3},{2,2}},{{3,-4},{-3,4}}}
    elsif ch='&' then
	path={{{-1,-4},{-2,-3},{-2,-2},{2,4}},{{-1,-4},{0,-4},{1,-3},{1,-2},{0,-1},{-2,0},{-3,2},{-3,3},{-2,4},{0,4},{2,2}}}
    elsif ch='\'' then
	path={{{-2,-4},{-2,-3},{-3,-2}}} dx=4
    elsif ch='(' then
	path={{{0,-4},{-1,-3},{-2,-1},{-2,2},{-1,4},{0,5}}} dx=5
    elsif ch=')' then
	path={{{-2,-4},{-1,-3},{0,-1},{0,2},{-1,4},{-2,5}}} dx=5
    elsif ch='*' then
	path={{{-2,-3},{2,3}},{{2,-3},{-2,3}},{{-3,0},{3,0}}}
    elsif ch='+' then
	path={{{0,-3},{0,3}},{{-3,0},{3,0}}}
    elsif ch=',' then
	path={{{-2,4},{-2,5},{-3,6}}} dx=3
    elsif ch='-' then
	path={{{-3,0},{3,0}}}
    elsif ch='.' then
	path={} dots={{-2,4}} dx=4
    elsif ch='/' then
	path={{{3,-4},{-3,4}}}
    elsif ch='0' then
	path={{{-1,-4},{1,-4},{3,-2},{3,2},{1,4},{-1,4},{-3,2},{-3,-2},{-1,-4}},{{3,-4},{-3,4}}}
    elsif ch='1' then
	path={{{-3,-3},{-2,-4},{-1,-4},{-1,4}},{{-3,4},{1,4}}} dx=5
    elsif ch='2' then
	path={{{-3,-3},{-2,-4},{2,-4},{3,-3},{3,-1},{2,0},{-1,0},{-3,2},{-3,4},{3,4}}}
    elsif ch='3' then
	path={{{-3,-3},{-2,-4},{2,-4},{3,-3},{3,-1},{2,0},{0,0}},{{2,0},{3,1},{3,3},{2,4},{-2,4},{-3,3}}}
    elsif ch='4' then
	path={{{1,-4},{-3,1},{-3,2},{2,2}},{{1,-4},{1,4}}}
    elsif ch='5' then
	path={{{3,-4},{-3,-4},{-3,-1},{2,-1},{3,0},{3,3},{2,4},{-2,4},{-3,3}}}
    elsif ch='6' then
	path={{{3,-3},{2,-4},{-2,-4},{-3,-3},{-3,3},{-2,4},{2,4},{3,3},{3,1},{2,0},{-2,0},{-3,1}}}
    elsif ch='7' then
	path={{{-3,-3},{-3,-4},{3,-4},{3,-2},{1,1},{0,4}}}
    elsif ch='8' then
	path={{{2,-4},{-2,-4},{-3,-3},{-3,-1},{-2,0},{2,0},{3,1},{3,3},{2,4},{-2,4},{-3,3},{-3,1},{-2,0}},{{2,0},{3,-1},{3,-3},{2,-4}}}
    elsif ch='9' then
	path={{{3,-1},{2,0},{-2,0},{-3,-1},{-3,-3},{-2,-4},{2,-4},{3,-3},{3,3},{2,4},{-2,4},{-3,3}}}
    elsif ch=':' then
	path={} dots={{-2,0},{-2,4}} dx=4
    elsif ch=';' then
	path={{{-2,3},{-2,4},{-3,5}}} dots={{-2,0}} dx=4
    elsif ch='<' then
	path={{{3,-3},{-3,0},{3,3}}}
    elsif ch='=' then
	path={{{-3,0},{3,0}},{{-3,2},{3,2}}}
    elsif ch='>' then
	path={{{-3,-3},{3,0},{-3,3}}}
    elsif ch='?' then
	path={{{-3,-3},{-2,-4},{1,-4},{2,-3},{2,-2},{0,0},{0,1}}} dots={{0,4}}
    elsif ch='@' then
	path={{{2,-1},{0,-2},{-2,-1},{-2,1},{0,2},{2,1},{2,-1}},{{2,1},{2,2},{3,1},{3,-2},{1,-4},{-1,-4},{-3,-2},{-3,2},{-1,4},{1,4}}}
    elsif ch='A' then
	path={{{-3,4},{-3,-1},{-1,-4},{1,-4},{3,-1},{3,4}},{{-3,1},{3,1}}}
    elsif ch='B' then
	path={{{-2,-4},{-2,4}},{{-3,-4},{2,-4},{3,-3},{3,-1},{2,0},{-2,0}},{{2,0},{3,1},{3,3},{2,4},{-3,4}}}
    elsif ch='C' then
	path={{{3,-3},{2,-4},{-2,-4},{-3,-3},{-3,3},{-2,4},{2,4},{3,3}}}
    elsif ch='D' then
	path={{{-2,-4},{-2,4}},{{-3,-4},{2,-4},{3,-3},{3,3},{2,4},{-3,4}}}
    elsif ch='E' then
	path={{{3,-4},{-3,-4},{-3,4},{3,4}},{{-3,0},{1,0}}}
    elsif ch='F' then
	path={{{3,-4},{-3,-4},{-3,4}},{{-3,0},{1,0}}}
    elsif ch='G' then
	path={{{3,-3},{2,-4},{-2,-4},{-3,-3},{-3,3},{-2,4},{2,4},{3,3},{3,0},{1,0}}}
    elsif ch='H' then
	path={{{-3,-4},{-3,4}},{{3,-4},{3,4}},{{-3,0},{3,0}}}
    elsif ch='I' then
	path={{{-3,-4},{1,-4}},{{-1,-4},{-1,4}},{{-3,4},{1,4}}} dx=6
    elsif ch='J' then
	path={{{-1,-4},{3,-4}},{{1,-4},{1,3},{0,4},{-2,4},{-3,3}}} dx=7
    elsif ch='K' then
	path={{{-3,-4},{-3,4}},{{3,-4},{-3,2}},{{-1,0},{3,4}}}
    elsif ch='L' then
	path={{{-3,-4},{-3,4},{3,4}}}
    elsif ch='M' then
	path={{{-3,4},{-3,-4},{0,0},{3,-4},{3,4}}}
    elsif ch='N' then
	path={{{-3,4},{-3,-4},{3,4},{3,-4}}}
    elsif ch='O' then
	path={{{2,-4},{-2,-4},{-3,-3},{-3,3},{-2,4},{2,4},{3,3},{3,-3},{2,-4}}}
    elsif ch='P' then
	path={{{-3,-4},{-3,4}},{{-3,-4},{2,-4},{3,-3},{3,-1},{2,0},{-3,0}}}
    elsif ch='Q' then
	path={{{2,-4},{-2,-4},{-3,-3},{-3,3},{-2,4},{2,4},{3,3},{3,-3},{2,-4}},{{1,3},{2,3},{3,4},{4,4}}} dx=9
    elsif ch='R' then
	path={{{-3,-4},{-3,4}},{{-3,-4},{2,-4},{3,-3},{3,-1},{2,0},{-3,0}},{{0,0},{3,4}}}
    elsif ch='S' then
	path={{{3,-3},{2,-4},{-2,-4},{-3,-3},{-3,-1},{-2,0},{2,0},{3,1},{3,3},{2,4},{-2,4},{-3,3}}}
    elsif ch='T' then
	path={{{-3,-4},{3,-4}},{{0,-4},{0,4}}}
    elsif ch='U' then
	path={{{-3,-4},{-3,3},{-2,4},{2,4},{3,3},{3,-4}}}
    elsif ch='V' then
	path={{{-3,-4},{0,4},{3,-4}}}
    elsif ch='W' then
	path={{{-4,-4},{-2,4},{0,-2},{2,4},{4,-4}}} dx=9
    elsif ch='X' then
	path={{{-3,-4},{3,4}},{{3,-4},{-3,4}}}
    elsif ch='Y' then
	path={{{-3,-4},{0,0},{3,-4}},{{0,0},{0,4}}}
    elsif ch='Z' then
	path={{{-3,-4},{3,-4},{-3,4},{3,4}}}
    elsif ch='[' then
	path={{{1,-4},{-2,-4},{-2,4},{1,4}}} dx=7
    elsif ch='\\' then
	path={{{-3,-4},{3,4}}}
    elsif ch=']' then
	path={{{-2,-4},{1,-4},{1,4},{-2,4}}} dx=7
    elsif ch='^' then
	path={{{-3,-1},{0,-4},{3,-1}}}
    elsif ch='_' then
	path={{{-3,5},{3,5}}}
    elsif ch='`' then
	path={{{-2,-4},{-2,-3},{-1,-2}}} dx=4
    elsif ch='a' then
	path={{{-2,-1},{1,-1},{2,0},{2,3},{3,4}},{{2,0},{1,1},{-2,1},{-3,2},{-3,3},{-2,4},{1,4},{2,3}}} accent=3
    elsif ch='b' then
	path={{{-3,-4},{-3,4}},{{-3,1},{-1,-1},{2,-1},{3,0},{3,3},{2,4},{-1,4},{-3,2}}}
    elsif ch='c' then
	path={{{3,0},{2,-1},{-2,-1},{-3,0},{-3,3},{-2,4},{2,4},{3,3}}} accent=3
    elsif ch='d' then
	path={{{3,1},{1,-1},{-2,-1},{-3,0},{-3,3},{-2,4},{1,4},{3,2}},{{3,-4},{3,4}}}
    elsif ch='e' then
	path={{{-3,1},{3,1},{3,0},{2,-1},{-2,-1},{-3,0},{-3,3},{-2,4},{2,4},{3,3}}} accent=3
    elsif ch='f' then
	path={{{2,-3},{1,-4},{0,-4},{-1,-3},{-1,4}},{{-3,0},{1,0}}} dx=6
    elsif ch='g' then
	path={{{3,1},{1,-1},{-2,-1},{-3,0},{-3,3},{-2,4},{1,4},{3,2}},{{3,-1},{3,6},{2,7},{-2,7},{-3,6}}} accent=3
    elsif ch='h' then
	path={{{-3,-4},{-3,4}},{{-3,1},{-1,-1},{2,-1},{3,0},{3,4}}}
    elsif ch='i' then
	path={{{-3,-1},{-2,-1},{-2,4}},{{-3,4},{-1,4}}} dots={{-2,-3}} dx=4 accent=3
    elsif ch='j' then
	path={{{-2,-1},{-1,-1},{-1,5},{-2,7},{-3,7},{-4,6}}} dots={{-1,-3}} dx=5 accent=3
    elsif ch='k' then
	path={{{-2,-4},{-2,4}},{{3,-1},{-2,3}},{{0,2},{3,4}}}
    elsif ch='l' then
	path={{{-3,-4},{-2,-4},{-2,4}},{{-3,4},{-1,4}}} dx=4 accent=5
    elsif ch='m' then
	path={{{-3,-1},{-3,4}},{{-3,0},{-2,-1},{-1,-1},{0,0},{0,4}},{{0,0},{1,-1},{2,-1},{3,0},{3,4}}} accent=3
    elsif ch='n' then
	path={{{-3,-1},{-3,4}},{{-3,1},{-1,-1},{2,-1},{3,0},{3,4}}} accent=3
    elsif ch='o' then
	path={{{2,-1},{-2,-1},{-3,0},{-3,3},{-2,4},{2,4},{3,3},{3,0},{2,-1}}} accent=3
    elsif ch='p' then
	path={{{-3,-1},{-3,7}},{{-3,1},{-1,-1},{2,-1},{3,0},{3,3},{2,4},{-1,4},{-3,2}}} accent=3
    elsif ch='q' then
	path={{{3,0},{1,-1},{-2,-1},{-3,0},{-3,3},{-2,4},{1,4},{3,2}},{{3,-1},{3,6},{4,7},{5,7},{6,6}}} dx=9 accent=3
    elsif ch='r' then
	path={{{-3,-1},{-3,4}},{{-3,1},{-1,-1},{2,-1},{3,0}}} accent=3
    elsif ch='s' then
	path={{{2,-1},{-2,-1},{-3,0},{-2,1},{2,1},{3,2},{3,3},{2,4},{-2,4},{-3,3}}} accent=3
    elsif ch='t' then
	path={{{-3,-1},{1,-1}},{{-1,-3},{-1,3},{0,4},{1,4},{2,3}}} dx=7 accent=1
    elsif ch='u' then
	path={{{-3,-1},{-3,3},{-2,4},{1,4},{3,2},{3,-1}},{{3,-1},{3,4}}} accent=3
    elsif ch='v' then
	path={{{-3,-1},{0,4},{3,-1}}} accent=3
    elsif ch='w' then
	path={{{-3,-1},{-2,4},{0,0},{2,4},{3,-1}}} accent=3
    elsif ch='x' then
	path={{{-3,-1},{3,4}},{{3,-1},{-3,4}}} accent=3
    elsif ch='y' then
	path={{{-3,-1},{0,4}},{{3,-1},{0,4},{-2,7},{-3,7}}} accent=3
    elsif ch='z' then
	path={{{-3,-1},{3,-1},{-3,4},{3,4}}} accent=3
    elsif ch='{' then
	path={{{2,-4},{1,-4},{0,-3},{0,-1},{-1,0},{0,1},{0,3},{1,4},{2,4}}}
    elsif ch='|' then
	path={{{-2,-4},{-2,-1}},{{-2,1},{-2,4}}} dx=4
    elsif ch='}' then
	path={{{-2,-4},{-1,-4},{0,-3},{0,-1},{1,0},{0,1},{0,3},{-1,4},{-2,4}}}
    elsif ch='~' then
	path={{{-3,-2},{-2,-3},{-1,-3},{1,-1},{2,-1},{3,-2}}}
    
    elsif ch=128 then --euro
	path={{{3,-3},{2,-4},{0,-4},{-2,-2},{-2,2},{0,4},{2,4},{3,3}},{{-3,-1},{1,-1}},{{-3,1},{1,1}}}
    elsif ch=136 then --identity
	path={{{-3,-1},{3,-1}},{{-3,1},{3,1}},{{-3,3},{3,3}}}
    elsif ch=137 then --congruent
	path={{{-3,-1},{-2,-2},{-1,-2},{1,0},{2,0},{3,-1}},{{-3,2},{-2,1},{-1,1},{1,3},{2,3},{3,2}}}
    elsif ch=138 then --le
	path={{{2,-2},{-3,0},{2,2}},{{-3,3},{2,3}}} dx=7
    elsif ch=139 then --ge
	path={{{-3,-2},{2,0},{-3,2}},{{-3,3},{2,3}}} dx=7
    elsif ch=140 then --times
	path={{{-3,-2},{2,3}},{{-3,3},{2,-2}}} dx=7
    elsif ch=141 then --divide
	path={{{-3,1},{3,1}}} dots={{0,-1},{0,3}}
    elsif ch=142 then --sqrt
	path={{{-3,1},{-2,1},{-0,4},{2,-4},{3,-4}}}
    elsif ch=143 then --neq
	path={{{-3,0},{3,0}},{{-3,2},{3,2}},{{2,-2},{-2,4}}}
    elsif ch=144 then --Delta
	path={{{-3,4},{0,-4},{3,4},{-3,4}}}
    elsif ch=145 then --Sigma
	path={{{3,-4},{-3,-4},{1,0},{-3,4},{3,4}}}
    elsif ch=146 then --Omega
	path={{{-3,4},{-1,4},{-2,2},{-3,1},{-3,-2},{-1,-4},{1,-4},{3,-2},{3,1},{2,2},{1,4},{3,4}}}
    elsif ch=147 then --alpha
	path={{{1,0},{0,-1},{-1,-1},{-3,1},{-3,3},{-2,4},{0,4},{1,3},{1,0},{2,-1},{3,-1}},{{1,3},{2,4},{3,4}}}
    elsif ch=148 then --beta
	path={{{-3,6},{-2,0},{-1,-3},{0,-4},{1,-4},{2,-3},{2,-2},{1,-1},{-1,-1}},{{0,-1},{1,0},{2,2},{2,3},{1,4},{-3,4}}} dx=7
    elsif ch=149 then --delta
	path={{{0,0},{-2,0},{-3,1},{-3,3},{-2,4},{0,4},{1,3},{1,1},{0,-2},{0,-3},{1,-4},{2,-4}}} dx=7
    elsif ch=150 then --epsilon
	path={{{2,-1},{1,-2},{-2,-2},{-3,-1},{-3,0},{-2,1},{-3,2},{-3,3},{-2,4},{1,4},{2,3}},{{-2,1},{1,1}}} dx=7
    elsif ch=151 then --theta
	path={{{0,-4},{-1,-4},{-2,-3},{-3,0},{-2,3},{-1,4},{0,4},{1,3},{2,0},{1,-3},{0,-4}},{{-3,0},{2,0}}} dx=7
    elsif ch=152 then --iota
	path={{{-2,-1},{-2,3},{-1,4},{0,4},{1,3}}} dx=6
    elsif ch=153 then --lambda
	path={{{-3,-4},{-2,-4},{-1,-3},{2,4}},{{0,0},{-3,4}}} dx=7
    elsif ch=154 then --mu
	path={{{-3,-1},{-3,7}},{{-3,2},{-1,4},{0,4},{2,2}},{{2,-1},{2,3},{3,4}}}
    elsif ch=155 then --pi
	path={{{-3,0},{-2,-1},{3,-1}},{{-1,-1},{-1,4}},{{2,-1},{2,4}}}
    elsif ch=156 then --rho
	path={{{-3,7},{-1,1},{0,0},{2,0},{3,2},{2,4},{0,4},{-1,3}}}     
    elsif ch=157 then --sigma
	path={{{1,0},{2,1},{2,3},{1,4},{-2,4},{-3,3},{-3,1},{-2,0},{3,-1}}}     
    elsif ch=158 then --phi
	path={{{2,-1},{-2,-1},{-3,0},{-3,3},{-2,4},{2,4},{3,3},{3,0},{2,-1}},{{1,-4},{-1,7}}}
    elsif ch=159 then --omega
	path={{{-2,0},{-3,1},{-3,3},{-2,4},{-1,4},{0,3},{0,1}},{{0,3},{1,4},{2,4},{3,3},{3,1},{2,0}}}    
    elsif ch=160 then --nbsp
	path={} dx=5
    elsif ch=161 then --inverted !
	path={{{-2,4},{-2,-1}}} dots={{-2,-4}} dx=4
    elsif ch=162 then --cent
	path={{{3,0},{2,-1},{-2,-1},{-3,0},{-3,3},{-2,4},{2,4},{3,3}},{{1,-3},{-1,6}}}     
    elsif ch=163 then --sterling
	path={{{2,-3},{1,-4},{0,-4},{-1,-3},{-1,3},{-2,4},{-3,3},{-2,2},{-1,2},{2,4},{3,3}},{{-3,-1},{1,-1}}}
    elsif ch=164 then --money
	path={{{-1,-2},{-2,-1},{-2,1},{-1,2},{1,2},{2,1},{2,-1},{1,-2},{-1,-2}},{{-2,-2},{-3,-3}},{{-2,2},{-3,3}},{{2,2},{3,3}},{{2,-2},{3,-3}}}
    elsif ch=165 then --yen
	path={{{-3,-4},{0,0},{3,-4}},{{0,0},{0,4}},{{-2,0},{2,0}},{{-2,2},{2,2}}}
    elsif ch=167 then --section
	path={{{0,-3},{-1,-4},{-2,-4},{-3,-3},{-3,-2},{-1,-1},{0,0},{-1,1},{-2,1}},{{-1,-1},{-2,-1},{-3,0},{-2,1},{0,2},{0,3},{-1,4},{-2,4},{-3,3}}}
    elsif ch=169 then --copy
	path={{{2,-2},{1,-3},{-1,-3},{-2,-2},{-2,0},{-1,1},{1,1},{2,0}},{{-2,-4},{-3,-3},{-3,1},{-2,2},{2,2},{3,1},{3,-3},{2,-4},{-2,-4}}}
    elsif ch=170 then --ordinal a
	path={{{-2,-4},{-1,-4},{0,-3},{0,-1}},{{0,-3},{-2,-3},{-3,-2},{-2,-1},{0,-1}},{{-3,0},{0,0}}} dx=5
    elsif ch=171 then --left guil
	path={{{-1,-1},{-3,1},{-1,3}},{{2,-1},{0,1},{2,3}}} dx=7
    elsif ch=173 then --trademark
	path={{{-3,-4},{-1,-4}},{{-2,-4},{-2,-1}},{{0,-1},{0,-4},{1,-2},{2,-4},{2,-1}}} dx=7
    elsif ch=174 then --reg
	path={{{-2,1},{-2,-3},{1,-3},{2,-2},{1,-1},{-2,-1}},{{0,-1},{2,1}},{{-2,-4},{-3,-3},{-3,1},{-2,2},{2,2},{3,1},{3,-3},{2,-4},{-2,-4}}}
    elsif ch=176 then --degree
	path={{{-2,-4},{-3,-3},{-3,-2},{-2,-1},{-1,-1},{0,-2},{0,-3},{-1,-4},{-2,-4}}} dx=5
    elsif ch=177 then --plusminus
	path={{{-3,0},{3,0}},{{0,-3},{0,3}},{{-3,4},{3,4}}}
    elsif ch=179 then --left dquo
	path={{{-2,-5},{-3,-4},{-3,-3},{-2,-2}},{{1,-5},{0,-4},{0,-3},{1,-2}}} dots={{-2,-2},{1,-2}} dx=7
    elsif ch=180 then --right dquo
	path={{{-2,-5},{-1,-4},{-1,-3},{-2,-2}},{{1,-5},{2,-4},{2,-3},{1,-2}}} dots={{-2,-5},{1,-5}} dx=7
    elsif ch=182 then --paragraph
	path={{{2,-4},{2,4}},{{0,-4},{0,4}},{{3,-4},{-2,-4},{-3,-3},{-3,-1},{-2,0},{2,0}}}
    elsif ch=183 then --raised dot
	path={} dots={{-2,0}} dx=5
    elsif ch=185 then --infinity
	path={{{-3,-1},{-3,1},{-2,2},{-1,2},{1,-2},{2,-2},{3,-1},{3,1},{2,2},{1,2},{-1,-2},{-2,-2},{-3,-1}}}
    elsif ch=186 then --ordinal o
	path={{{-2,-4},{-3,-3},{-3,-2},{-2,-1},{-1,-1},{0,-2},{0,-3},{-1,-4},{-2,-4}},{{-3,0},{0,0}}} dx=5
    elsif ch=187 then --right guil
	path={{{-3,-1},{-1,1},{-3,3}},{{0,-1},{2,1},{0,3}}} dx=7
    elsif ch=188 then --1/4
	path={{{-3,-4},{-3,-1}},{{1,-3},{-3,3}},{{2,0},{0,3},{3,3}},{{2,0},{2,4}}}
    elsif ch=189 then --1/2
	path={{{-3,-4},{-3,-1}},{{1,-3},{-3,3}},{{0,1},{1,0},{2,0},{3,1},{3,2},{1,3},{0,4},{3,4}}}
    elsif ch=190 then --3/4
	path={{{-3,-4},{-2,-4},{-1,-3},{-2,-2},{-3,-2}},{{-2,-2},{-1,-1},{-2,0},{-3,0}},{{2,-3},{-2,3}},{{2,0},{0,3},{3,3}},{{2,0},{2,4}}}
    elsif ch=191 then --inverted ?
	path={{{-1,-1},{-1,0},{-3,2},{-3,3},{-2,4},{1,4},{2,3}}} dots={{-1,-4}} dx=7
    elsif ch=192 then --A grave
	path={{{-3,4},{-3,-1},{-1,-4},{1,-4},{3,-1},{3,4}},{{-3,1},{3,1}},{{-2,-7},{1,-6}}}
    elsif ch=193 then --A acute
	path={{{-3,4},{-3,-1},{-1,-4},{1,-4},{3,-1},{3,4}},{{-3,1},{3,1}},{{2,-7},{-1,-6}}}
    elsif ch=194 then --A cir
	path={{{-3,4},{-3,-1},{-1,-4},{1,-4},{3,-1},{3,4}},{{-3,1},{3,1}},{{-2,-6},{0,-8},{2,-6}}}
    elsif ch=195 then --C cir
	path={{{3,-3},{2,-4},{-2,-4},{-3,-3},{-3,3},{-2,4},{2,4},{3,3}},{{-2,-6},{0,-8},{2,-6}}}
    elsif ch=196 then --A uml
	path={{{-3,4},{-3,-1},{-1,-4},{1,-4},{3,-1},{3,4}},{{-3,1},{3,1}}} dots={{-1,-6},{1,-6}}
    elsif ch=197 then --G cir
	path={{{3,-3},{2,-4},{-2,-4},{-3,-3},{-3,3},{-2,4},{2,4},{3,3},{3,0},{1,0}},{{-2,-6},{0,-8},{2,-6}}}
    elsif ch=198 then --AE
	path={{{-3,4},{-3,-1},{-1,-4},{1,-4},{1,4}},{{-3,1},{1,1}},{{1,-4},{3,-4}},{{1,0},{3,0}},{{1,4},{3,4}}}
    elsif ch=199 then --C cedilla
	path={{{3,-3},{2,-4},{-2,-4},{-3,-3},{-3,3},{-2,4},{2,4},{3,3}},{{0,4},{0,5},{1,5},{0,6},{-1,6}}}
    elsif ch=200 then --E grave
	path={{{3,-4},{-3,-4},{-3,4},{3,4}},{{-3,0},{1,0}},{{-2,-7},{1,-6}}}
    elsif ch=201 then --E acute
	path={{{3,-4},{-3,-4},{-3,4},{3,4}},{{-3,0},{1,0}},{{2,-7},{-1,-6}}}
    elsif ch=202 then --E cir
	path={{{3,-4},{-3,-4},{-3,4},{3,4}},{{-3,0},{1,0}},{{-2,-6},{0,-8},{2,-6}}}
    elsif ch=203 then --E uml
	path={{{3,-4},{-3,-4},{-3,4},{3,4}},{{-3,0},{1,0}}} dots={{-1,-6},{1,-6}}
    elsif ch=204 then --I grave
	path={{{-3,-4},{1,-4}},{{-1,-4},{-1,4}},{{-3,4},{1,4}},{{-3,-7},{0,-6}}} dx=6
    elsif ch=205 then --I acute
	path={{{-3,-4},{1,-4}},{{-1,-4},{-1,4}},{{-3,4},{1,4}},{{1,-7},{-2,-6}}} dx=6
    elsif ch=206 then --I cir
	path={{{-3,-4},{1,-4}},{{-1,-4},{-1,4}},{{-3,4},{1,4}},{{-3,-6},{-1,-8},{1,-6}}} dx=6
    elsif ch=207 then --I uml
	path={{{-3,-4},{1,-4}},{{-1,-4},{-1,4}},{{-3,4},{1,4}}} dots={{-2,-6},{0,-6}} dx=6
    elsif ch=208 then --H cir
	path={{{-3,-4},{-3,4}},{{3,-4},{3,4}},{{-3,0},{3,0}},{{-2,-6},{0,-8},{2,-6}}}
    elsif ch=209 then --N tilde
	path={{{-3,4},{-3,-4},{3,4},{3,-4}},{{-3,-6},{-2,-7},{-1,-7},{1,-5},{2,-5},{3,-6}}}
    elsif ch=210 then --O grave
	path={{{2,-4},{-2,-4},{-3,-3},{-3,3},{-2,4},{2,4},{3,3},{3,-3},{2,-4}},{{-2,-7},{1,-6}}}
    elsif ch=211 then --O acute
	path={{{2,-4},{-2,-4},{-3,-3},{-3,3},{-2,4},{2,4},{3,3},{3,-3},{2,-4}},{{2,-7},{-1,-6}}}
    elsif ch=212 then --O cir
	path={{{2,-4},{-2,-4},{-3,-3},{-3,3},{-2,4},{2,4},{3,3},{3,-3},{2,-4}},{{-2,-6},{0,-8},{2,-6}}}
    elsif ch=214 then --O uml
	path={{{2,-4},{-2,-4},{-3,-3},{-3,3},{-2,4},{2,4},{3,3},{3,-3},{2,-4}}} dots={{-1,-6},{1,-6}}
    elsif ch=215 then --OE
	path={{{1,-3},{0,-4},{-2,-4},{-3,-3},{-3,3},{-2,4},{0,4},{1,3}},{{1,-4},{1,4}},{{1,-4},{3,-4}},{{1,0},{3,0}},{{1,4},{3,4}}}
    elsif ch=216 then --S cir
	path={{{3,-3},{2,-4},{-2,-4},{-3,-3},{-3,-1},{-2,0},{2,0},{3,1},{3,3},{2,4},{-2,4},{-3,3}},{{-2,-6},{0,-8},{2,-6}}}
    elsif ch=217 then --U grave
	path={{{-3,-4},{-3,3},{-2,4},{2,4},{3,3},{3,-4}},{{-2,-7},{1,-6}}}
    elsif ch=218 then --U acute
	path={{{-3,-4},{-3,3},{-2,4},{2,4},{3,3},{3,-4}},{{2,-7},{-1,-6}}}
    elsif ch=219 then --U cir
	path={{{-3,-4},{-3,3},{-2,4},{2,4},{3,3},{3,-4}},{{-2,-6},{0,-8},{2,-6}}}
    elsif ch=220 then --U uml
	path={{{-3,-4},{-3,3},{-2,4},{2,4},{3,3},{3,-4}}} dots={{-1,-6},{1,-6}}
    elsif ch=221 then --U breve
	path={{{-3,-4},{-3,3},{-2,4},{2,4},{3,3},{3,-4}},{{-2,-6},{-1,-5},{1,-5},{2,-6}}}
    elsif ch=222 then --Y acute
	path={{{-3,-4},{0,0},{3,-4}},{{0,0},{0,4}},{{2,-7},{-1,-6}}}
    elsif ch=223 then --Y uml
	--path={{{-3,-4},{0,0},{3,-4}},{{0,0},{0,4}}} dots={{-1,-6},{1,-6}}
	path={{{-3,4},{-3,-3},{-2,-4},{0,-4},{1,-3},{1,-1},{0,0},{-2,0}},{{0,0},{1,1},{2,2},{2,3},{1,4},{-2,4}}} dx=7
    elsif ch=224 then --a grave
	path={{{-2,-1},{1,-1},{2,0},{2,3},{3,4}},{{2,0},{1,1},{-2,1},{-3,2},{-3,3},{-2,4},{1,4},{2,3}},{{-2,-4},{1,-3}}}
    elsif ch=225 then --a acute
	path={{{-2,-1},{1,-1},{2,0},{2,3},{3,4}},{{2,0},{1,1},{-2,1},{-3,2},{-3,3},{-2,4},{1,4},{2,3}},{{2,-4},{-1,-3}}}
    elsif ch=226 then --a cir
	path={{{-2,-1},{1,-1},{2,0},{2,3},{3,4}},{{2,0},{1,1},{-2,1},{-3,2},{-3,3},{-2,4},{1,4},{2,3}},{{-2,-3},{0,-5},{2,-3}}}
    elsif ch=227 then --c cir
	path={{{3,0},{2,-1},{-2,-1},{-3,0},{-3,3},{-2,4},{2,4},{3,3}},{{-2,-2},{0,-4},{2,-2}}}
    elsif ch=228 then --a uml
	path={{{-2,-1},{1,-1},{2,0},{2,3},{3,4}},{{2,0},{1,1},{-2,1},{-3,2},{-3,3},{-2,4},{1,4},{2,3}}} dots={{-1,-4},{1,-4}}
    elsif ch=229 then --g cir
	path={{{3,1},{1,-1},{-2,-1},{-3,0},{-3,3},{-2,4},{1,4},{3,2}},{{3,-1},{3,6},{2,7},{-2,7},{-3,6}},{{-2,-3},{0,-5},{2,-3}}}
    elsif ch=230 then --ae
	path={{{-3,-1},{-1,-1},{0,0},{0,3}},{{0,0},{-1,1},{-2,1},{-3,2},{-3,3},{-2,4},{0,3},{0,4}},{{0,1},{3,1},{3,0},{2,-1},{1,-1},{0,0},{0,3},{1,4},{2,4},{3,3}}}
    elsif ch=231 then --c cedilla
	path={{{3,0},{2,-1},{-2,-1},{-3,0},{-3,3},{-2,4},{2,4},{3,3}},{{0,4},{0,5},{1,5},{0,6},{-1,6}}}
    elsif ch=232 then --e grave
	path={{{-3,1},{3,1},{3,0},{2,-1},{-2,-1},{-3,0},{-3,3},{-2,4},{2,4},{3,3}},{{-2,-4},{1,-3}}}
    elsif ch=233 then --e acute
	path={{{-3,1},{3,1},{3,0},{2,-1},{-2,-1},{-3,0},{-3,3},{-2,4},{2,4},{3,3}},{{2,-4},{-1,-3}}}
    elsif ch=234 then --e cir
	path={{{-3,1},{3,1},{3,0},{2,-1},{-2,-1},{-3,0},{-3,3},{-2,4},{2,4},{3,3}},{{-2,-3},{0,-5},{2,-3}}}
    elsif ch=235 then --e uml
	path={{{-3,1},{3,1},{3,0},{2,-1},{-2,-1},{-3,0},{-3,3},{-2,4},{2,4},{3,3}}} dots={{-1,-3},{1,-3}}
    elsif ch=236 then --i grave
	path={{{-2,-1},{-1,-1},{-1,4}},{{-2,4},{0,4}},{{-3,-4},{0,-3}}} dx=5
    elsif ch=237 then --i acute
	path={{{-2,-1},{-1,-1},{-1,4}},{{-2,4},{0,4}},{{1,-4},{-2,-3}}} dx=5
    elsif ch=238 then --i cir
	path={{{-2,-1},{-1,-1},{-1,4}},{{-2,4},{0,4}},{{-3,-3},{-1,-5},{1,-3}}} dx=5
    elsif ch=239 then --i uml
	path={{{-2,-1},{-1,-1},{-1,4}},{{-2,4},{0,4}}} dots={{-2,-3},{0,-3}} dx=5
    elsif ch=240 then --h cir
	path={{{-3,-4},{-3,4}},{{-3,1},{-1,-1},{2,-1},{3,0},{3,4}},{{-2,-5},{0,-7},{2,-5}}}
    elsif ch=241 then --n tilde
	path={{{-3,-1},{-3,4}},{{-3,1},{-1,-1},{2,-1},{3,0},{3,4}},{{-2,-3},{-1,-4},{0,-4},{1,-3},{2,-3},{3,-4}}}
    elsif ch=242 then --o grave
	path={{{2,-1},{-2,-1},{-3,0},{-3,3},{-2,4},{2,4},{3,3},{3,0},{2,-1}},{{-2,-4},{1,-3}}}
    elsif ch=243 then --o acute
	path={{{2,-1},{-2,-1},{-3,0},{-3,3},{-2,4},{2,4},{3,3},{3,0},{2,-1}},{{2,-4},{-1,-3}}}
    elsif ch=244 then --o cir
	path={{{2,-1},{-2,-1},{-3,0},{-3,3},{-2,4},{2,4},{3,3},{3,0},{2,-1}},{{-2,-3},{0,-5},{2,-3}}}
    elsif ch=246 then --o uml
	path={{{2,-1},{-2,-1},{-3,0},{-3,3},{-2,4},{2,4},{3,3},{3,0},{2,-1}}} dots={{-1,-3},{1,-3}}
    elsif ch=247 then --oe
	path={{{0,0},{-1,-1},{-2,-1},{-3,0},{-3,3},{-2,4},{-1,4},{0,3}},{{0,1},{3,1},{3,0},{2,-1},{1,-1},{0,0},{0,3},{1,4},{2,4},{3,3}}}
    elsif ch=248 then --s cir
	path={{{2,-1},{-2,-1},{-3,0},{-2,1},{2,1},{3,2},{3,3},{2,4},{-2,4},{-3,3}},{{-2,-3},{0,-5},{2,-3}}}
    elsif ch=249 then --u grave
	path={{{-3,-1},{-3,3},{-2,4},{1,4},{3,2},{3,-1}},{{3,-1},{3,4}},{{-2,-4},{1,-3}}}
    elsif ch=250 then --u acute
	path={{{-3,-1},{-3,3},{-2,4},{1,4},{3,2},{3,-1}},{{3,-1},{3,4}},{{2,-4},{-1,-3}}}
    elsif ch=251 then --u cir
	path={{{-3,-1},{-3,3},{-2,4},{1,4},{3,2},{3,-1}},{{3,-1},{3,4}},{{-2,-3},{0,-5},{2,-3}}}
    elsif ch=252 then --u uml
	path={{{-3,-1},{-3,3},{-2,4},{1,4},{3,2},{3,-1}},{{3,-1},{3,4}}} dots={{-1,-3},{1,-3}}
    elsif ch=253 then --u breve
	path={{{-3,-1},{-3,3},{-2,4},{1,4},{3,2},{3,-1}},{{3,-1},{3,4}},{{-2,-4},{-1,-3},{1,-3},{2,-4}}}
    elsif ch=254 then --y acute
	path={{{-3,-1},{0,4}},{{3,-1},{0,4},{-2,7},{-3,7}},{{2,-4},{-1,-3}}}
    elsif ch=255 then --y uml
	path={{{-3,-1},{0,4}},{{3,-1},{0,4},{-2,7},{-3,7}}} dots={{-1,-3},{1,-3}}
--accents for use with put_accent()
    elsif ch=256 then --grave
	path={{{-2,-7},{1,-6}}}
    elsif ch=257 then --acute
	path={{{2,-7},{-1,-6}}}
    elsif ch=258 then --circumflex
	path={{{-2,-6},{0,-8},{2,-6}}}
    elsif ch=259 then --umlaut
	path={} dots={{-1,-6},{1,-6}}
    elsif ch=260 then --hacek
	path={{{-2,-8},{0,-6},{2,-8}}}
    elsif ch=261 then --tilde
	path={{{-2,-6},{-1,-7},{0,-7},{1,-6},{2,-6},{3,-7}}}
    elsif ch=262 then --breve
	path={{{-2,-7},{-1,-6},{1,-6},{2,-7}}}
    elsif ch=263 then --circle
	path={{{0,-5},{1,-5},{2,-6},{2,-7},{1,-8},{0,-8},{-1,-7},{-1,-6},{0,-5}}}
    elsif ch=264 then --long
	path={{{-2,-6},{2,-6}}}
    elsif ch=265 then --double acute
	path={{{0,-8},{-1,-6}},{{2,-8},{1,-6}}}
    elsif ch=266 then --dot
	path={} dots={{0,-6}}
    elsif ch=267 then --hook
	path={{{0,-8},{-1,-7},{-1,-6},{0,-6}}} dots={{0,-6}}
    elsif ch=268 then --cedilla
	path={{{0,4},{0,5},{1,5},{0,6},{-1,6}}}
    elsif ch=269 then --base right hook
	path={{{1,5},{0,6},{0,7},{1,7}}}
    elsif ch=270 then --base left hook
	path={{{0,5},{1,5},{1,6},{0,7}}}
    elsif ch=271 then
	path={} dots={{0,6}}
    end if
    if length(path) then
	draw_modified_path(path,pos)
	if bold then
	    pos[1]+=1
	    draw_modified_path(path,pos)
	    pos[2]+=1
	    draw_modified_path(path,pos)
	    pos[1]-=1
	    draw_modified_path(path,pos)
	    pos[2]-=1
	end if
    end if
    if length(dots) then
	draw_dots(dots,pos)
	if bold then
	    pos[1]+=1
	    draw_dots(dots,pos)
	    pos[2]+=1
	    draw_dots(dots,pos)
	    pos[1]-=1
	    draw_dots(dots,pos)
	    pos[2]-=1
	end if
    end if
end procedure

procedure get_char_width(integer char)
    --get width of one character; originally returned from graphics_char (dx)
    width=8
    if char=',' then
	width=3
    elsif char='!' or char='\'' or char='.' or char='`' or char=161 then
	width=4
    elsif char=':' or char=';' or char='i' or char='l' or char='|' then
	width=4
    elsif char='(' or char=')' or char='1' or char='j' or char=' ' then
	width=5
    elsif char=160 or char=170 or char=176 or char=183 then
	width=5
    elsif char=186 or (char>=236 and char<=239) then
	width=5
    elsif char='\"' or char='I' or char='f' or char=152 then
	width=6
    elsif char=170 or char=186 or (char>=204 and char<=207) then
	width=6
    elsif char='J' or char='[' or char=']' or char='t' or char=128 then
	width=7
    elsif (char>=138 and char<=140) or (char>=148 and char<=151) then
	width=7
    elsif char=153 or char=171 or char=173 or char=179 then
	width=7
    elsif char=180 or char=187 or char=191 or char=223 then
	width=7
    elsif char='Q' or char='W' or char='q'then
	width=9
    end if
end procedure

function text_width(sequence text)
    --pixel width of string (proportional chars)
    integer total
    total=0
    for n=1 to length(text) do
	get_char_width(text[n])
	total+=width
    end for
    return total
end function

global procedure graphics_text(sequence text,sequence pos,integer just,
	    integer pro)
    --draw text at pos ({x,y}, either 0 to 1 or actual pixels); start at pos
    --(left justify) if just>0, center if just=0, end at pos if just<0;
    --fixed width of 8*scale if pro=0, otherwise proportional
    atom total,x,y
    integer adj
    graphics_info()
    x=pos[1] y=pos[2]
    if x<=1 then
	x=x*hres
    end if
    if y<=1 then
	y=y*vres
    end if
    end_y=y
    if pro then
	total=text_width(text)*scale
    else
	total=8*length(text)*scale
    end if
    if just=0 then --centered
	end_x=x+total/2
	x=floor(x-(total/2)+4*scale)
    elsif just>0 then --left justification
	end_x=x+total
	x=floor(x+4*scale)
    elsif just<0 then --right justification
	end_x=x
	x=floor(x-total+4*scale)
    end if
    for n=1 to length(text) do
	get_char_width(text[n])
	if pro then
	    graphics_char(text[n],{x,y})
	    x+=width*scale
	else
	    adj=0
	    if width<7 then
		adj+=1
	    end if
	    if width<5 then
		adj+=1
	    end if
	    graphics_char(text[n],{x+scale*adj,y})
	    x+=8*scale
	    width=8
	end if
    end for
end procedure

global procedure continue_text(sequence text,integer pro)
    --display text starting at end of preceding text
    graphics_text(text,{end_x,end_y},1,pro)
end procedure

global procedure put_accent(integer accent_type)
    --add accent to previous character; type can be 1 to 16
    if accent_type>0 and accent_type<=16 then
	if accent_type>11 then
	    accent=0 --baseline accents
	end if
	graphics_char(accent_type+255,{end_x-(scale*width/2),end_y+accent*scale})
    end if
end procedure

global procedure show_chars()
    --show all defined character at scale=2; characters are left justified in
    --7x9 matrix; will be centered (fixed) or use fewer pixels (proportional)
    --when displayed with graphics_text()
    text_attributes(2,8,0,0)--scale=2, color=black, not bold, not italics
    for r=2 to 15 do
	for n=0 to 15 do
	    graphics_char(16*r+n,{10*n*scale+30,14*r*scale+20})
	end for
    end for
end procedure

